#!/usr/bin/env python

"""
Handlers for a person for whom scheduling is performed, inspecting outgoing
messages to obtain scheduling done externally.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from imiptools.content import Handler
from imiptools.data import get_window_end, uri_dict, uri_item, uri_values
from imiptools.period import remove_affected_period

class PersonHandler(Handler):

    "Handling mechanisms specific to people."

    def _get_identity(self, from_organiser=True):

        """
        Get the identity of interest in a usable form for any unprocessed
        object.
        """

        identity, attr = item = uri_item(self.obj.get_item(from_organiser and "ORGANIZER" or "ATTENDEE"))

        # Check for event using UID.

        if not self.have_new_object(identity):
            return None

        return item

    def _record(self, from_organiser=True, update_freebusy=False):

        "Record free/busy and object information."

        item = self._get_identity(from_organiser)
        if not item:
            return False

        identity, attr = item

        # Update the object.

        if from_organiser:

            # Set the complete event or an additional occurrence.

            self.store.set_event(identity, self.uid, self.recurrenceid, self.obj.to_node())

            # Remove additional recurrences if handling a complete event.

            if not self.recurrenceid:
                self.store.remove_recurrences(identity, self.uid)

        else:
            organiser_item, attendees = self.require_organiser_and_attendees(from_organiser)
            self.merge_attendance(attendees, identity)

        # Remove any associated request.

        self.store.dequeue_request(identity, self.uid, self.recurrenceid)

        # Update free/busy information.

        if update_freebusy:

            freebusy = self.store.get_freebusy(identity)

            # Interpretation of periods can depend on the time zone.

            tzid = self.get_tzid(identity)

            # Use the stored event in case the reply is incomplete, as is seen
            # when Claws sends a REPLY for an object originally employing
            # recurrence information.

            obj = self.get_object(identity)

            # If newer than any old version, discard old details from the
            # free/busy record and check for suitability.

            periods = obj.get_periods_for_freebusy(tzid, get_window_end(tzid))

            self.update_freebusy_for_participant(freebusy, periods, attr,
                from_organiser and self.is_not_attendee(identity, obj))

            # Remove either original recurrence or additional recurrence
            # details depending on whether an additional recurrence or a
            # complete event are being handled, respectively.

            self.remove_freebusy_for_recurrences(freebusy, self.store.get_recurrences(identity, self.uid))
            self.store.set_freebusy(identity, freebusy)

            if self.publisher:
                self.publisher.set_freebusy(identity, freebusy)

        return True

    def _remove(self, from_organiser=True, update_freebusy=False):

        "Remove free/busy information for any unprocessed object."

        item = self._get_identity(from_organiser)
        if not item:
            return False

        identity, attr = item

        # Only cancel the event completely if all attendees are given.
        # NOTE: Need to also check for recurrence identifiers and selective
        # NOTE: cancellations.

        obj = self.get_object(identity)

        attendees = uri_dict(obj.get_value_map("ATTENDEE"))
        all_attendees = set(attendees.keys())
        given_attendees = set(uri_values(self.obj.get_values("ATTENDEE")))

        if given_attendees == all_attendees:
            self.store.cancel_event(identity, self.uid, self.recurrenceid)

        # Otherwise, remove the given attendees and update the event.

        else:
            for attendee in given_attendees:
                if attendees.has_key(attendee):
                    del attendees[attendee]
            obj["ATTENDEE"] = attendees.items()

        # Update the stored object with sequence information.

        obj["SEQUENCE"] = self.obj.get_items("SEQUENCE")
        obj["DTSTAMP"] = self.obj.get_items("DTSTAMP")

        # Set the complete event if not an additional occurrence.

        self.store.set_event(identity, self.uid, self.recurrenceid, obj.to_node())

        # Remove any associated request.

        self.store.dequeue_request(identity, self.uid, self.recurrenceid)

        # Update free/busy information.

        if update_freebusy:
            freebusy = self.store.get_freebusy(identity)
            self.remove_from_freebusy(freebusy)
            self.store.set_freebusy(identity, freebusy)

            if self.publisher:
                self.publisher.set_freebusy(identity, freebusy)

        return True

class Event(PersonHandler):

    "An event handler."

    def add(self):
        pass

    def cancel(self):
        self._remove(True, True)

    def counter(self):
        pass

    def declinecounter(self):
        pass

    def publish(self):
        self._record(True, True)

    def refresh(self):
        pass

    def reply(self):
        self._record(False, True)

    def request(self):
        self._record(True, True)

class Freebusy(PersonHandler):

    "A free/busy handler."

    def publish(self):
        pass

    def reply(self):
        pass

    def request(self):
        pass

class Journal(PersonHandler):

    "A journal entry handler."

    def add(self):
        pass

    def cancel(self):
        self._remove(True)

    def publish(self):
        self._record(True)

class Todo(PersonHandler):

    "A to-do item handler."

    def add(self):
        pass

    def cancel(self):
        self._remove(True)

    def counter(self):
        pass

    def declinecounter(self):
        pass

    def publish(self):
        self._record(True)

    def refresh(self):
        pass

    def reply(self):
        self._record(False)

    def request(self):
        self._record(True)

# Handler registry.

handlers = [
    ("VFREEBUSY",   Freebusy),
    ("VEVENT",      Event),
    ("VTODO",       Todo),
    ("VJOURNAL",    Journal),
    ]

# vim: tabstop=4 expandtab shiftwidth=4
